/*
 * Lange Nacht der Wissenschaften 2022
 *
 * Gedächtnistrainer
 *
 * 30.6.2022  V1.0  Sven Tschirley
 *
 */

#include <msp430g2313.h>
//#include <msp430g2001.h>
#include <stdlib.h>

/****************************************************/

#define LED_R    BIT0
#define LED_G    BIT1
#define LED_B    BIT2
#define LED_Y    BIT3
#define LED_ALL  (LED_R|LED_G|LED_B|LED_Y)
#define BUTTON_R BIT4
#define BUTTON_G BIT5
#define BUTTON_B BIT6
#define BUTTON_Y BIT7
#define BUTTON_ALL (BUTTON_R|BUTTON_G|BUTTON_B|BUTTON_Y)

#define MAX_STEPS   16        // das kann man ja auch erhöhen
#define MAX_TIMEOUT 10
#define MAX_UNUSED   4

// Zustände des Automaten
enum
  {
    MODE_GAME_START,
    MODE_GAME_SIMON,
    MODE_GAME_PLAYER,
    MODE_GAME_END_FAIL,
    MODE_GAME_END_WIN,
    MODE_GAME_SLEEP
  };

enum
  {
    LS_START,
    LS_GAME_END_FAIL,
    LS_GAME_END_WIN,
    LS_SLEEP_START
  };

/****************************************************/
unsigned char steps[MAX_STEPS+1]; // Speicher
unsigned char currentStep = 0;    // gibt an, in welchem Schritt das Spiel ist
unsigned char stepCtr     = 0;    // Zählt Spielereingaben
unsigned int  timeoutCtr  = 0;    // Zählt für Spieler-Timeout
unsigned char unusedCtr   = 0;    // Zählt Durchläufe ohne Bedienung
unsigned char buttonIn    = 0;    // Zwischenspeicher für Taste
unsigned char random;             // Zufallszahl zwischen 1 und 4
unsigned char idx;                // Zählvariable
         int  delayDuration;      // Leuchtdauer
unsigned char gameState   = MODE_GAME_START; // Zustandsspeicher

volatile unsigned int lfsr = 0xACE1;
unsigned long int seed = 0xA1ED034;
/****************************************************/
/*
 * Funktionsprototypen
 */

/****************************************************/
void          ConfigWDT( void );
void          ConfigClocks( void );
void          ConfigPorts( void );
void          ConfigTimerA0( void );
void          IlluminateLED( unsigned char color, unsigned int duration);
unsigned char getRandom( void );
void          delay_ms( int msec );

/*
 * Konfiguration des Watchdogs -- einfach ausschalten,
 * den brauchen wir nicht. Der Controller arbeitet auch
 * mit fast garkeiner Batteriespannung
 */
void ConfigWDT( void )
{
  WDTCTL = WDTPW + WDTHOLD; // Stop WDT
}

/*
 * Konfigurieren der Takterzeugung
 */
void ConfigClocks( void )
{
  BCSCTL1 = CALBC1_1MHZ;            // Set range
  DCOCTL = CALDCO_1MHZ;             // Set DCO step + modulation
  BCSCTL3 |= LFXT1S_2;                      // LFXT1 = VLO
  IFG1 &= ~OFIFG;                           // Clear OSCFault flag
  BCSCTL2 |= SELM_0 + DIVM_3 + DIVS_3;      // MCLK = DCO/8, SMCLK = DCO/8
}

/*
 * Konfigurieren des Port1
 * Alle LEDs sind Ausgänge
 * Alle Buttons sind Eingänge
 */
void ConfigPorts( void )
{
  P1DIR = 0;            // All unused pins are output (SLAU144J, sec 8.2.8
  P2DIR = 0;            // All unused pins are output (SLAU144J, sec 8.2.8
  /* Initialisierung der LEDs                                   */
  P1DIR |= LED_ALL;     // Alle LED pins als Ausgang schalten
  P1OUT &= LED_ALL;     // Alle LEDs ausschalten
  /* Eingänge                                                   */
  P1DIR &= ~BUTTON_ALL; // Buttons sind Eingaenge
  P1REN |= BUTTON_ALL;  // Pull-Up/Down einschalten
  P1OUT |= BUTTON_ALL;  // Pull-Up auswählen
  /* Interruptsteuerung                                         */
  P1IE  |= BUTTON_ALL;  // P1.4 bis P1.7 interrupt enabled
  P1IES |= BUTTON_ALL;  // P1.4 bis p1.7 Hi/lo edge
  P1IFG &= ~BUTTON_ALL; // IFG cleared

  //P2DIR |= BIT0;        // Pin für Scope-Debugging
}

/*
 * Konfigurieren des Timers für Delay
 * Takt 1MHZ, Quelle ist SMCLK, CCR wird
 */
void ConfigTimerA0( void )
{
  TACCTL1 = OUTMOD_0;
  TACCR1  = 32;
  TACCTL0 = CCIE;            // Interrupt wird hier nicht gebraucht
  TACCR0  = 0x0100;          // setzt delay-Funktion neu
  TA0CTL  = TASSEL_2 + MC_2; // Source SMCLK, Cont.Mode
}

/****************************************************/
/*
 * Interrupt Service Routinen (ISR)
 */

/*
 * nicht verwentete Interrupts werden als leere Funktionen
 * implementiert, damit nicht irgendwas eingefangen wird.
 * Das sollte nicht passieren, da die alle ausgeschaltet sind,
 * aber man kann ja nie wissen...
 */
#pragma vector=PORT1_VECTOR
__interrupt void Port_1_ISR( void ){
  P1IFG &= ~BUTTON_ALL;  // IFG löschen
  buttonIn = (~P1IN) & BUTTON_ALL; // Eingabe speichern
  delayDuration++;
}

//#pragma vector=ADC10_VECTOR
//__interrupt void ADC10_ISR( void ){}

#pragma vector=PORT2_VECTOR
__interrupt void Port_2_ISR( void ){}

#pragma vector=NMI_VECTOR
__interrupt void NMI_ISR( void ){}

#pragma vector=TIMER0_A0_VECTOR
__interrupt void Timer_A (void)
{
  // wake on CCR0 count match
  TA0CCTL0 = 0;
  //__bic_SR_register_on_exit(LPM0_bits|GIE);
}

#pragma vector=TIMER0_A1_VECTOR
__interrupt void Timer_A1_ISR( void ){}

//#pragma vector=USI_VECTOR
//__interrupt void USI_ISR( void ){}

#pragma vector=WDT_VECTOR
__interrupt void WDT_ISR( void ){}

/****************************************************/
/* Relevante Funktionen für den Programmablauf      */
/*
 * Einschalten einer LED für eine übergebene Dauer
 */
void IlluminateLED( unsigned char color, unsigned int duration)
{
  P1OUT |= color;
  delay_ms( duration );
  P1OUT &= ~color;
}

/*
 * Initialisierung für Zufallsgenerator errechnen
 */
unsigned long int rnd31 ( void )
{
    unsigned long int hi, lo;
    lo = 16807 * (seed & 0x0FFFF);
    hi = 16807 * (seed >> 16 );
    lo += (hi & 0x7FFF) << 16;
    lo +=  hi >> 15;
    if ( lo > 0x7FFFFFFF )
        lo -= 0x7FFFFFFF;
    seed = lo;
    return ( lo );
}

/*
 * Erzeugen einer Zufallszahl aus 1..4
 */
unsigned char getRandom( void )
{
  unsigned char tmp;

  tmp = ( rand() & 0x03 ) + 1;
  return( tmp );
}

/*
 * Füllen der Sequenz
 */
void fillSteps( void )
{
  unsigned int iseed = lfsr;

  /* Zufallsgenerator initialisieren                            */
  seed = rnd31();
  iseed = ( (unsigned int) seed );
  srand( iseed );
  // Zufall einmal laufenlassen
  for( idx = 0; idx<MAX_STEPS; idx++)
    {
      steps[idx] = getRandom();
    }
}

void Lightshow ( unsigned char mode )
{
    switch ( mode)
      {
      case LS_START:
        IlluminateLED( LED_R, 100 );
        IlluminateLED( LED_G, 100 );
        IlluminateLED( LED_B, 100 );
        IlluminateLED( LED_Y, 100 );
        IlluminateLED( LED_B, 100 );
        IlluminateLED( LED_G, 100 );
        IlluminateLED( LED_R, 100 );
        break;
      case  LS_GAME_END_FAIL:
        IlluminateLED( LED_ALL, 1000 );
        delay_ms( 1000 );
        IlluminateLED( LED_ALL, 1000 );
        delay_ms( 1000 );
        break;
      case LS_GAME_END_WIN:
        IlluminateLED( LED_R, 100 );
        IlluminateLED( LED_B, 100 );
        IlluminateLED( LED_Y, 100 );
        IlluminateLED( LED_G, 100 );
        IlluminateLED( LED_Y, 100 );
        IlluminateLED( LED_G, 100 );
        IlluminateLED( LED_R, 100 );
        IlluminateLED( LED_B, 100 );
        break;
      case LS_SLEEP_START:
        IlluminateLED( LED_Y, 300 );
        IlluminateLED( LED_B, 300 );
        IlluminateLED( LED_G, 300 );
        IlluminateLED( LED_R, 300 );
        break;
      default:
        break;
      }

}

/*
 * Timer-basierte Verzögerung
 * DCO ist 1 MHz, Timertakt 125kHz -> T = 8us
 * als CCR Wert 125, damit ca. 1 ms delay erzeugt wird,
 * so gut der interne Oszillator das kann
 */
void delay_ms(int  msec)
{
  int i = 0;
  // change according to your SMCLK frequency
  TA0CCR0 = 125;
  TA0CTL = TASSEL_2 + MC_2;
  for( i = 0; i < msec; ++i )
    {
      TA0R = 0;
      TA0CCTL0 = CCIE;
      while( TA0R < 80 )
        {
          __no_operation();
        }
    }
  TA0CTL = 0;
}

/****************************************************/
/*
 * Hauptprogramm
 */
int main ( void )
{
  /* Aufruf der Initialisierungsfunktion fuer den Controller    */
  ConfigWDT();      // Watchdog
  ConfigClocks();   // Takt
  ConfigPorts();    // LEDs
  ConfigTimerA0();  // Timer

  fillSteps();      // Array mit Zufallsfarben füllen

  _BIS_SR( GIE );

  /* Endlosschleife mit Spiel                                   */
  while ( 1 ) // oder while( gameState != MODE_SLEEP )
    {
      switch( gameState )
        {
        case MODE_GAME_START:
          Lightshow( LS_START );
          fillSteps();      // Array mit Zufallsfarben füllen
          buttonIn    = 0;  // Eingabe-Puffer löschen
          currentStep = 0;
          stepCtr     = 0;
          timeoutCtr  = 0;
          delayDuration = 500;
          delay_ms( 500 );      // Pause vor Start
          gameState = MODE_GAME_SIMON;
          break;
          // Abschnitt für Ausgabe der Sequenz durch den Controller aus
          // dem Speicher bis zum erreichen von MAX_STEPS
        case MODE_GAME_SIMON:
          random = getRandom();
          // Zufallszahl der Farbe zuordnen und diese speichern
          if      ( random == 1 ) steps[currentStep++] = LED_R;
          else if ( random == 2 ) steps[currentStep++] = LED_G;
          else if ( random == 3 ) steps[currentStep++] = LED_B;
          else                    steps[currentStep++] = LED_Y;
          // Maximale Schrittzahl erreicht?
          if ( currentStep >= MAX_STEPS )
            {
              gameState = MODE_GAME_END_WIN;
            }
          // Inhalt des Speichers wiedergeben
          for ( idx=0; idx < currentStep; idx++ )
            {
              IlluminateLED( steps[ idx ], delayDuration );
              delay_ms( delayDuration );
            }
          // Zähler für Spieler zurücksetzen
          stepCtr = 0;
          timeoutCtr = 0;
          delay_ms( 500 );       // Pause vor dem PLAYER
          // Spieler-Modus folgt, falls wir nicht schon gewonnen haben
          if ( gameState != MODE_GAME_END_WIN )
            {
              gameState = MODE_GAME_PLAYER;
            }
          break;
          // Spieler muss Sequenz wiedergeben
          //
        case MODE_GAME_PLAYER:
          while ( ( stepCtr < currentStep ) && ( gameState != MODE_GAME_END_FAIL ) )
            {
              // solange das Ende nicht erreicht wird läuft diese Schleife,
              // die Eingaben werden in der ISR in buttonIn hineingeschrieben
              // die Durchläufe werden auf 500 ms verzögert
              if ( buttonIn != 0 )
                {
                  // Button RED
                  if ( buttonIn == BUTTON_R )
                    {
                      IlluminateLED( LED_R, 500 );
                      if ( steps[stepCtr] != LED_R )
                        {
                          gameState = MODE_GAME_END_FAIL;
                        }
                    }
                  // Button GREEN
                  if ( buttonIn == BUTTON_G )
                    {
                      IlluminateLED( LED_G, 500 );
                      if ( steps[stepCtr] != LED_G )
                        {
                          gameState = MODE_GAME_END_FAIL;
                        }
                    }
                  // Button BLUE
                  if ( buttonIn == BUTTON_B )
                    {
                      IlluminateLED( LED_B, 500 );
                      if ( steps[stepCtr] != LED_B )
                        {
                          gameState = MODE_GAME_END_FAIL;
                        }
                    }
                  // Button YELLOW
                  if ( buttonIn == BUTTON_Y )
                    {
                      IlluminateLED( LED_Y, 500 );
                      if ( steps[stepCtr] != LED_Y )
                        {
                          gameState = MODE_GAME_END_FAIL;
                        }
                    }
                  buttonIn = 0;
                  // nächsten Durchlauf vorbereiten
                  timeoutCtr = 0;
                  stepCtr++;
                } // if (buttonIn != 0 )
              else
                { // keine Eingabe, warte für 5 sec, dann haben wir verloren
                  timeoutCtr++;
                  if ( timeoutCtr == MAX_TIMEOUT )
                    {
                      gameState = MODE_GAME_END_FAIL;
                    }
                  else
                    {
                      delay_ms( 500 );
                    }
                }
            } // while ( stepCtr < currentStep )
          delay_ms( 500 );   // Pause, dann ist SIMON dran
          // Spieler-Modus folgt, falls wir nicht schon verloren haben
          if ( gameState != MODE_GAME_END_FAIL )
            {
              gameState = MODE_GAME_SIMON;
            }
          break;
        case MODE_GAME_END_FAIL:
          if ( stepCtr == 0)
            {
              unusedCtr++;
            }
          Lightshow( LS_GAME_END_FAIL );
          if ( unusedCtr == 0x03 )
            {
              Lightshow( LS_SLEEP_START );
              gameState = MODE_GAME_SLEEP;
            }
           else
            {
              gameState = MODE_GAME_START;
            }
          break;
        case MODE_GAME_END_WIN:
          for( idx=0; idx <= 5; idx++ )
            {
              Lightshow( LS_GAME_END_WIN );
            }
          delay_ms( 1000 );
          gameState = MODE_GAME_START;
          break;
          // Schlafen bis Taste gedrückt wird, dann starten
        case MODE_GAME_SLEEP:
          P1OUT &= ~LED_ALL;
          if( buttonIn != 0 )
            {
              unusedCtr = 0;
              buttonIn  = 0;
              gameState = MODE_GAME_START;
            }
        default:
          break;
        }
      //IlluminateLED( LED_B, 1000);
      _delay_cycles( 1000 );
    } // ende von  while ( 1 )
  return( 0 );

}

/* EOF */




